/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.component.business.dao.hibernate.interceptor;

import org.hibernate.EmptyInterceptor;
import org.hibernate.Interceptor;
import org.hibernate.type.Type;
import org.openvpms.component.business.dao.hibernate.cache.ArchetypeIdCache;
import org.openvpms.component.business.domain.archetype.ArchetypeId;

import java.io.Serializable;

/**
 * An {@link Interceptor} that replaces {@link ArchetypeId} instances with those cached by an {@link ArchetypeIdCache},
 * to avoid proliferation of {@link ArchetypeId} instances.
 *
 * @author Tim Anderson
 */
public class IMObjectInterceptor extends EmptyInterceptor {

    /**
     * The ArchetypeId cache.
     */
    private final ArchetypeIdCache cache;

    /**
     * Constructs an {@link IMObjectInterceptor}.
     */
    public IMObjectInterceptor() {
        this(new ArchetypeIdCache());
    }

    /**
     * Constructs an {@link IMObjectInterceptor}.
     *
     * @param cache the ArchetypeId cache to use
     */
    public IMObjectInterceptor(ArchetypeIdCache cache) {
        this.cache = cache;
    }

    /**
     * Called just before an object is initialized. The interceptor may change the {@code state}, which will
     * be propagated to the persistent object. Note that when this method is called, {@code entity} will be
     * an empty uninitialized instance of the class.
     *
     * @param entity        The entity instance being loaded
     * @param id            The identifier value being loaded
     * @param state         The entity state (which will be pushed into the entity instance)
     * @param propertyNames The names of the entity properties, corresponding to the {@code state}.
     * @param types         The types of the entity properties, corresponding to the {@code state}.
     * @return {@code true} if the user modified the {@code state} in any way.
     */
    @Override
    public boolean onLoad(Object entity, Serializable id, Object[] state, String[] propertyNames, Type[] types) {
        boolean modified = false;
        for (int i = 0; i < state.length; ++i) {
            Object object = state[i];
            if (object instanceof ArchetypeId) {
                state[i] = cache.get((ArchetypeId) object);
                modified = true;
            }
        }
        return modified;
    }

}
