/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.plugin.internal.service.mapping;

import org.openvpms.component.model.archetype.ArchetypeDescriptor;
import org.openvpms.component.model.entity.Entity;
import org.openvpms.component.model.object.IMObject;
import org.openvpms.component.service.archetype.ArchetypeService;
import org.openvpms.mapping.exception.MappingException;
import org.openvpms.mapping.model.Mappings;
import org.openvpms.mapping.model.Targets;
import org.openvpms.mapping.service.MappingService;
import org.openvpms.plugin.service.i18n.MappingMessages;

import java.util.Comparator;
import java.util.List;

/**
 * Default implementation of {@link MappingServiceImpl}.
 *
 * @author Tim Anderson
 */
public class MappingServiceImpl implements MappingService {

    /**
     * The archetype service.
     */
    private final ArchetypeService service;

    /**
     * Constructs a {@link MappingServiceImpl}.
     *
     * @param service the archetype service
     */
    public MappingServiceImpl(ArchetypeService service) {
        this.service = service;
    }

    /**
     * Create a configuration to record mappings between the specified archetype and a target.
     *
     * @param type the type being mapped
     * @return a new configuration to record mappings
     */
    @Override
    public <T extends IMObject> IMObject createMappingConfiguration(Class<T> type) {
        if (!Entity.class.isAssignableFrom(type)) {
            throw new MappingException(MappingMessages.unsupportedType(type));
        }
        return service.create(EntityMappings.MAPPING);
    }

    /**
     * Creates mappings.
     *
     * @param config      the mapping configuration
     * @param type        the type being mapped
     * @param archetype   the archetype being mapped
     * @param displayName the mapping display name
     * @param targets     the targets
     * @return the mappings
     */
    @Override
    @SuppressWarnings("unchecked")
    public <T extends IMObject> Mappings createMappings(IMObject config, Class<T> type, String archetype,
                                                        String displayName, Targets targets) {
        List<ArchetypeDescriptor> descriptors = checkArchetype(archetype, type);
        String sourceDisplayName;
        if (descriptors.size() > 1) {
            descriptors.sort(Comparator.comparing(ArchetypeDescriptor::getDisplayName));
            sourceDisplayName = descriptors.get(0).getDisplayName();
            sourceDisplayName += ",...";
        } else {
            sourceDisplayName = descriptors.get(0).getDisplayName();
        }
        return new EntityMappings(config, type, archetype, displayName, sourceDisplayName, targets, service);
    }

    /**
     * Creates mappings.
     *
     * @param config            the mapping configuration
     * @param type              the type being mapped
     * @param archetype         the archetype being mapped
     * @param displayName       the mapping display name
     * @param sourceDisplayName the source object type display name
     * @param targets           the targets
     * @return the mappings
     */
    @Override
    @SuppressWarnings("unchecked")
    public <T extends IMObject> Mappings<T> createMappings(IMObject config, Class<T> type, String archetype,
                                                           String displayName, String sourceDisplayName,
                                                           Targets targets) {
        checkArchetype(archetype, type);
        return new EntityMappings(config, type, archetype, displayName, sourceDisplayName, targets, service);
    }

    /**
     * Checks that an archetype is valid.
     *
     * @param archetype the archetype. May contain wildcards
     * @param type      the expected implementation class
     * @return the matching archetype descriptors
     */
    private <T extends IMObject> List<ArchetypeDescriptor> checkArchetype(String archetype, Class<T> type) {
        if (!Entity.class.isAssignableFrom(type)) {
            throw new MappingException(MappingMessages.unsupportedType(type));
        }
        List<ArchetypeDescriptor> descriptors = service.getArchetypeDescriptors(archetype);
        if (descriptors.isEmpty()) {
            throw new MappingException(MappingMessages.archetypeNotFound(archetype));
        }
        for (ArchetypeDescriptor descriptor : descriptors) {
            if (!type.isAssignableFrom(descriptor.getClassType())) {
                throw new MappingException(MappingMessages.archetypeNotType(descriptor.getArchetypeType(), type));
            }
        }
        return descriptors;
    }
}
