/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.admin.mapping;

import nextapp.echo2.app.Component;
import nextapp.echo2.app.Table;
import nextapp.echo2.app.event.TableModelEvent;
import nextapp.echo2.app.event.TableModelListener;
import nextapp.echo2.app.table.DefaultTableColumnModel;
import nextapp.echo2.app.table.TableColumn;
import org.openvpms.component.model.object.IMObject;
import org.openvpms.component.system.common.query.SortConstraint;
import org.openvpms.mapping.model.Mapping;
import org.openvpms.mapping.model.Mappings;
import org.openvpms.web.component.im.layout.LayoutContext;
import org.openvpms.web.component.im.table.AbstractIMTableModel;
import org.openvpms.web.component.im.table.PagedIMTable;
import org.openvpms.web.component.property.AbstractSaveableEditor;
import org.openvpms.web.component.property.Validator;
import org.openvpms.web.echo.factory.ColumnFactory;
import org.openvpms.web.echo.style.Styles;
import org.openvpms.web.echo.table.DefaultTableCellRenderer;
import org.openvpms.web.echo.table.TableColumnFactory;

import java.util.List;

/**
 * .
 *
 * @author Tim Anderson
 */
public class MappingEditors<T extends IMObject> extends AbstractSaveableEditor {

    private final Mappings<T> mappings;

    private final LayoutContext context;

    private PagedIMTable<Mapping> table;

    private Component component;

    public MappingEditors(Mappings<T> mappings, LayoutContext context) {
        this.mappings = mappings;
        this.context = context;
    }

    /**
     * Save any edits.
     */
    @Override
    protected void doSave() {
        mappings.save();
    }

    /**
     * Validates the object.
     *
     * @param validator the validator
     * @return {@code true} if the object and its descendants are valid otherwise {@code false}
     */
    @Override
    protected boolean doValidation(Validator validator) {
        // TODO
        return true;
    }

    public Component getComponent() {
        if (component == null) {
            table = new PagedIMTable<>(new MappingEditorTableModel<>(mappings, context));
            style(table.getTable());
            refreshTable();
            component = ColumnFactory.create(Styles.INSET, table.getComponent());
            table.getTable().getModel().addTableModelListener((TableModelListener) event -> {
                if (event.getType() == TableModelEvent.UPDATE
                    && event.getFirstRow() == event.getLastRow()) {
                    setModified();
                    refreshTable();
                }
            });
        }
        return component;
    }

    private void style(Table table) {
        table.setStyleName("list");
        table.setDefaultRenderer(Object.class, DefaultTableCellRenderer.INSTANCE);
        table.setRolloverEnabled(false);
        table.setSelectionEnabled(false);
    }

    private void refreshTable() {
        int page = table.getModel().getPage();
        table.setResultSet(new MappingResultSet(null, 20, mappings));
        if (page != 0) {
            table.getNavigator().setPage(page);
        }
    }

    private static class MappingEditorTableModel<O extends IMObject> extends AbstractIMTableModel<Mapping> {

        private final Mappings<O> mappings;

        private final LayoutContext context;

        private MappingEditor[] editors;

        static final int SOURCE_INDEX = 0;

        static final int TARGET_INDEX = 1;

        MappingEditorTableModel(Mappings<O> mappings, LayoutContext context) {
            this.mappings = mappings;
            this.context = context;
            DefaultTableColumnModel model = new DefaultTableColumnModel();
            model.addColumn(TableColumnFactory.create(SOURCE_INDEX, mappings.getSourceDisplayName()));
            model.addColumn(TableColumnFactory.create(TARGET_INDEX, mappings.getTargetDisplayName()));
            setTableColumnModel(model);
        }

        /**
         * Sets the objects to display.
         *
         * @param objects the objects to display
         */
        @Override
        public void setObjects(List<Mapping> objects) {
            super.setObjects(objects);
            editors = new MappingEditor[objects.size()];
        }

        /**
         * Returns the sort criteria.
         *
         * @param column    the primary sort column
         * @param ascending if {@code true} sort in ascending order; otherwise sort in {@code descending} order
         * @return the sort criteria, or {@code null} if the column isn't sortable
         */
        @Override
        public SortConstraint[] getSortConstraints(int column, boolean ascending) {
            return null;
        }

        /**
         * Returns the value found at the given coordinate within the table.
         *
         * @param object the object
         * @param column the column
         * @param row    the row
         * @return the value at the given coordinate.
         */
        @Override
        protected Object getValue(Mapping object, TableColumn column, int row) {
            MappingEditor editor = editors[row];
            if (editor == null) {
                editor = new MappingEditor<>((MappingState) object, mappings, context);
                editor.addModifiableListener(modifiable -> fireTableRowsUpdated(row, row));
                editors[row] = editor;
            }
            if (column.getModelIndex() == SOURCE_INDEX) {
                return editor.getSource().getComponent();
            }
            return editor.getTarget().getComponent();
        }
    }
}
