/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.admin.organisation;

import echopointng.TabbedPane;
import nextapp.echo2.app.Component;
import nextapp.echo2.app.Label;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.mapping.model.Mappings;
import org.openvpms.mapping.service.MappingProvider;
import org.openvpms.plugin.manager.PluginManager;
import org.openvpms.plugin.service.config.ConfigurableService;
import org.openvpms.web.component.error.ErrorFormatter;
import org.openvpms.web.component.im.edit.AbstractIMObjectEditor;
import org.openvpms.web.component.im.edit.IMObjectEditor;
import org.openvpms.web.component.im.layout.AbstractLayoutStrategy;
import org.openvpms.web.component.im.layout.IMObjectLayoutStrategy;
import org.openvpms.web.component.im.layout.IMObjectTabPane;
import org.openvpms.web.component.im.layout.IMObjectTabPaneModel;
import org.openvpms.web.component.im.layout.LayoutContext;
import org.openvpms.web.component.property.Property;
import org.openvpms.web.echo.factory.ColumnFactory;
import org.openvpms.web.echo.factory.LabelFactory;
import org.openvpms.web.echo.factory.TabbedPaneFactory;
import org.openvpms.web.echo.style.Styles;
import org.openvpms.web.echo.tabpane.TabPaneModel;
import org.openvpms.web.system.ServiceHelper;
import org.openvpms.web.workspace.admin.mapping.MappingEditors;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * An editor for <em>entity.plugin*</em> plugin configurations.
 * <p/>
 * This supports plugins that implement the {@link MappingProvider} interface; a mapping tab will
 * be added when the plugin is configured.
 *
 * @author Tim Anderson
 */
public class PluginEditor extends AbstractIMObjectEditor {

    /**
     * The mapping editors.
     */
    private List<MappingEditors<?>> mappingEditors = new ArrayList<>();

    /**
     * The mapping container.
     */
    private Component mappingContainer;

    /**
     * The mapping tab pane.
     */
    private TabbedPane pane;

    /**
     * The logger.
     */
    private static final Logger log = LoggerFactory.getLogger(PluginEditor.class);

    /**
     * Constructs an {@link PluginEditor}.
     *
     * @param object        the object to edit
     * @param parent        the parent object. May be {@code null}
     * @param layoutContext the layout context
     */
    public PluginEditor(IMObject object, IMObject parent, LayoutContext layoutContext) {
        super(object, parent, layoutContext);
        mappingContainer = ColumnFactory.create(Styles.INSET_Y);
    }

    /**
     * Creates a new instance of the editor, with the latest instance of the object to edit.
     *
     * @return {@code null}
     */
    @Override
    public IMObjectEditor newInstance() {
        return new PluginEditor(reload(getObject()), getParent(), getLayoutContext());
    }

    /**
     * Save any edits.
     */
    @Override
    protected void doSave() {
        super.doSave();
        refresh();
    }

    /**
     * Refreshes the mapping display, if the plugin implements the {@link MappingProvider} interface.
     */
    protected void refresh() {
        int index = (pane != null) ? pane.getSelectedIndex() : -1;
        mappingContainer.removeAll();
        for (MappingEditors editors : mappingEditors) {
            getEditors().remove(editors);
        }
        mappingEditors.clear();

        try {
            IMObject config = getObject();
            if (!config.isNew()) {
                MappingProvider provider = getMappingProvider(config);
                if (provider != null) {
                    TabPaneModel model = new TabPaneModel(mappingContainer);

                    for (Mappings<?> mappings : provider.getMappings()) {
                        MappingEditors<?> editors = new MappingEditors<>(mappings, getLayoutContext());
                        mappingEditors.add(editors);
                        addEditor(editors);
                        model.addTab(mappings.getDisplayName(), editors.getComponent());
                    }
                    pane = TabbedPaneFactory.create(model);
                    if (index != -1 && index < pane.getModel().size()) {
                        pane.setSelectedIndex(index);
                    }
                    mappingContainer.add(pane);
                }
            }
        } catch (Throwable exception) {
            Label label = LabelFactory.create(true, true);
            label.setText(ErrorFormatter.format(exception));
            mappingContainer.add(label);
        }
    }

    /**
     * Invoked when layout has completed.
     * <p>
     * This can be used to perform processing that requires all editors to be created.
     */
    @Override
    protected void onLayoutCompleted() {
        super.onLayoutCompleted();
        if (!getObject().isNew()) {
            refresh();
        }
    }

    /**
     * Creates the layout strategy.
     *
     * @return a new layout strategy
     */
    @Override
    protected IMObjectLayoutStrategy createLayoutStrategy() {
        return new LayoutStrategy();
    }

    /**
     * Determines if the plugin configuration is associated with a {@link MappingProvider}.
     *
     * @param config the plugin configuration
     * @return the mapping provider, or {@code null} if the service doesn't implement the {@link MappingProvider}
     * interface
     */
    private MappingProvider getMappingProvider(IMObject config) {
        MappingProvider result = null;
        try {
            PluginManager manager = ServiceHelper.getBean(PluginManager.class);
            for (ConfigurableService service : manager.getServices(ConfigurableService.class)) {
                if (service instanceof MappingProvider) {
                    if (configures(config, service)) {
                        result = (MappingProvider) service;
                        break;
                    }
                }
            }
        } catch (Throwable exception) {
            log.warn(exception.getMessage(), exception);
        }
        return result;
    }

    /**
     * Determines if a configuration is used to configure the specified service.
     *
     * @param config  the plugin configuration
     * @param service the configuration service
     * @return {@code true} if the configuration configures the service
     */
    private boolean configures(IMObject config, ConfigurableService service) {
        boolean result = false;
        try {
            org.openvpms.component.model.object.IMObject configuration = service.getConfiguration();
            result = (configuration != null && configuration.equals(config));
        } catch (Throwable exception) {
            log.warn(exception.getMessage(), exception);
        }
        return result;
    }

    private class LayoutStrategy extends AbstractLayoutStrategy {

        /**
         * Lays out each child component in a tabbed pane.
         *
         * @param object     the object to lay out
         * @param parent     the parent object. May be {@code null}
         * @param properties the properties
         * @param container  the container to use
         * @param context    the layout context
         */
        @Override
        protected void doComplexLayout(IMObject object, IMObject parent, List<Property> properties, Component container,
                                       LayoutContext context) {
            IMObjectTabPaneModel model = doTabLayout(object, properties, container, context, false);
            model.addTab("Mappings", mappingContainer);
            IMObjectTabPane pane = new IMObjectTabPane(model);
            pane.setSelectedIndex(0);
            container.add(pane);
        }
    }
}
