/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.admin.system.diagnostics;

import nextapp.echo2.app.Button;
import nextapp.echo2.app.Component;
import nextapp.echo2.app.Label;
import nextapp.echo2.app.Row;
import nextapp.echo2.app.SelectField;
import nextapp.echo2.app.SplitPane;
import nextapp.echo2.app.event.ActionEvent;
import org.apache.log4j.Appender;
import org.apache.log4j.Logger;
import org.apache.log4j.RollingFileAppender;
import org.openvpms.component.business.domain.im.document.Document;
import org.openvpms.web.component.bound.BoundTextComponentFactory;
import org.openvpms.web.component.property.ModifiableListener;
import org.openvpms.web.component.property.Property;
import org.openvpms.web.component.property.SimpleProperty;
import org.openvpms.web.component.util.ErrorHelper;
import org.openvpms.web.echo.event.ActionListener;
import org.openvpms.web.echo.factory.ButtonFactory;
import org.openvpms.web.echo.factory.ColumnFactory;
import org.openvpms.web.echo.factory.LabelFactory;
import org.openvpms.web.echo.factory.RowFactory;
import org.openvpms.web.echo.factory.SelectFieldFactory;
import org.openvpms.web.echo.factory.SplitPaneFactory;
import org.openvpms.web.echo.style.Styles;
import org.openvpms.web.echo.text.TextField;
import org.openvpms.web.resource.i18n.format.NumberFormatter;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPOutputStream;

/**
 * Log viewer.
 *
 * @author Tim Anderson
 */
public class LogViewer extends AbstractDiagnosticTab {

    /**
     * Page selector.
     */
    private final Property pageSelector;

    /**
     * Page selector listener.
     */
    private final ModifiableListener pageListener;

    /**
     * The viewer pane.
     */
    private final SplitPane pane;

    /**
     * The current reader.
     */
    private LogReader reader;

    /**
     * The current page.
     */
    private int page = 0;

    /**
     * First page.
     */
    private Button first;

    /**
     * Previous page.
     */
    private Button previous;

    /**
     * Next page.
     */
    private Button next;

    /**
     * Last page.
     */
    private Button last;

    /**
     * Log file size.
     */
    private Label size;

    /**
     * Page input.
     */
    private TextField input;

    /**
     * Maximum number of lines per page.
     */
    private static final int MAX_LINES = 100;

    /**
     * Constructs a {@link LogViewer}.
     */
    LogViewer() {
        super("admin.system.diagnostic.log");
        pane = SplitPaneFactory.create(SplitPane.ORIENTATION_VERTICAL_TOP_BOTTOM, "LogViewer");

        pageSelector = new SimpleProperty("page", this.page, Integer.class);
        pageListener = modifiable -> show(pageSelector.getInt() - 1);
        pageSelector.addModifiableListener(pageListener);
    }

    /**
     * Returns the log file paths.
     *
     * @return the paths
     */
    public List<String> getFiles() {
        List<String> files = new ArrayList<>();
        String fileName = getDefaultFileName();
        if (fileName != null) {
            if (Files.isReadable(Paths.get(fileName))) {
                files.add(fileName);
            }
        }
        boolean done = false;
        int suffix = 1;
        while (!done) {
            String path = fileName + "." + suffix;
            if (Files.isReadable(Paths.get(path))) {
                files.add(path);
                suffix++;
            } else {
                done = true;
            }
        }
        return files;
    }

    /**
     * Returns a document containing the diagnostics.
     *
     * @return the document, or {@code null} if one cannot be created
     */
    @Override
    public Document getDocument() {
        Document result = null;
        LogReader reader = getReader();
        if (reader != null) {
            result = getDocument(reader.getPath());
        }
        return result;
    }

    /**
     * Returns a document containing the diagnostics for the specified file.
     *
     * @param file the log file
     * @return the document, or {@code null} if one cannot be created
     */
    public Document getDocument(String file) {
        Document result = null;
        try {
            byte[] buffer = new byte[1024];
            ByteArrayOutputStream bytes = new ByteArrayOutputStream();
            GZIPOutputStream gzip = new GZIPOutputStream(bytes);
            FileInputStream in = new FileInputStream(file);

            int count;
            while ((count = in.read(buffer)) > 0) {
                gzip.write(buffer, 0, count);
            }

            in.close();

            gzip.finish();
            gzip.close();
            byte[] content = bytes.toByteArray();
            result = toDocument(file + ".gz", content, content.length, "application/gzip");
        } catch (Throwable exception) {
            ErrorHelper.show(exception);
        }
        return result;
    }

    /**
     * Returns the diagnostic content.
     *
     * @return the diagnostic content, or {@code null} if it cannot be generated
     */
    @Override
    protected Component getContent() {
        pane.removeAll();
        List<String> files = getFiles();

        first = ButtonFactory.create(
                null, "navigation.first", new ActionListener() {
                    public void onAction(ActionEvent event) {
                        onFirst();
                    }
                });

        input = BoundTextComponentFactory.create(pageSelector, 5);
        previous = ButtonFactory.create(
                null, "navigation.previous", new ActionListener() {
                    public void onAction(ActionEvent event) {
                        onPrevious();
                    }
                });
        next = ButtonFactory.create(
                null, "navigation.next", new ActionListener() {
                    public void onAction(ActionEvent event) {
                        onNext();
                    }
                });
        last = ButtonFactory.create(
                null, "navigation.last", new ActionListener() {
                    public void onAction(ActionEvent event) {
                        onLast();
                    }
                });
        size = LabelFactory.create();

        Component log;

        if (!files.isEmpty()) {
            read(files.get(0));
            SelectField fileSelector = SelectFieldFactory.create(files.toArray());
            fileSelector.addActionListener(new ActionListener() {
                @Override
                public void onAction(ActionEvent event) {
                    Object selectedItem = fileSelector.getSelectedItem();
                    if (selectedItem instanceof String) {
                        read(selectedItem.toString());
                        show(0);
                    }
                }
            });
            log = RowFactory.create(Styles.CELL_SPACING, fileSelector, size);
        } else {
            log = LabelFactory.create("admin.system.diagnostic.lognotfound");
        }
        enablePaging(reader != null);

        Row row = RowFactory.create(Styles.CELL_SPACING, first, previous, input, next, last);
        pane.add(ColumnFactory.create(Styles.INSET, ColumnFactory.create(Styles.CELL_SPACING, log, row)));

        show(this.page);
        return pane;
    }

    /**
     * Sets the file name to read,
     *
     * @param fileName the file name.
     */
    private void read(String fileName) {
        reader = new LogReader(fileName, MAX_LINES);
    }

    /**
     * Returns the reader.
     *
     * @return the reader
     */
    private LogReader getReader() {
        if (reader == null) {
            List<String> paths = getFiles();
            if (!paths.isEmpty()) {
                read(paths.get(0));
            }
        }
        return reader;
    }

    /**
     * Enables/disables paging.
     *
     * @param enable if {@code true} enable the buttons, otherwise disable them
     */
    private void enablePaging(boolean enable) {
        first.setEnabled(enable);
        last.setEnabled(enable);
        input.setEnabled(enable);
        previous.setEnabled(enable);
        next.setEnabled(enable);
    }

    /**
     * Returns the default file name.
     *
     * @return the default file name
     */
    private String getDefaultFileName() {
        String fileName = null;
        Appender appender = Logger.getRootLogger().getAppender("fullout");
        if (appender instanceof RollingFileAppender) {
            fileName = ((RollingFileAppender) appender).getFile();
        }
        return fileName;
    }

    /**
     * Displays the first page.
     */
    private void onFirst() {
        page = show(0);
    }

    /**
     * Displays the previous page.
     */
    private void onPrevious() {
        if (page > 0) {
            page--;
        }
        page = show(page);
    }

    /**
     * Displays the next page.
     */
    private void onNext() {
        page = show(page + 1);
    }

    /**
     * Displays the last page.
     */
    private void onLast() {
        LogReader.Page content = null;
        try {
            content = reader.readLast();
            if (content == null) {
                // truncated. Try and re-read
                content = reader.readLast();
            }
        } catch (IOException exception) {
            ErrorHelper.show(exception);
        }
        page = show(content);
    }

    /**
     * Show a page.
     *
     * @param page the page number (0-based)
     * @return the actual page number
     */
    private int show(int page) {
        LogReader.Page content = null;
        try {
            content = reader.read(page);
            if (content == null) {
                // page does not exist. Try and show the last page
                content = reader.readLast();
            }
        } catch (IOException exception) {
            ErrorHelper.show(exception);
        }
        return show(content);
    }

    /**
     * Shows page content.
     *
     * @param content the content. May be {@code null}
     * @return the current page
     */
    private int show(LogReader.Page content) {
        if (pane.getComponentCount() == 2) {
            pane.remove(1);
        }
        size.setText(NumberFormatter.getSize(getLength()));
        String buffer = "";
        if (content != null) {
            StringBuilder builder = new StringBuilder();
            int number = content.getFirstLine();
            for (String line : content.getLines()) {
                builder.append(++number).append(": ").append(line).append('\n');
            }
            buffer = builder.toString();
        }
        pageSelector.removeModifiableListener(pageListener);
        pageSelector.setValue(content != null ? content.getPage() + 1 : null);
        pageSelector.addModifiableListener(pageListener);
        pane.add(ColumnFactory.create(Styles.INSET, LabelFactory.preformatted(buffer)));
        return (content != null) ? content.getPage() : 0;
    }

    /**
     * Returns the log file length.
     *
     * @return the log file length
     */
    private long getLength() {
        try {
            return reader.getLength();
        } catch (Throwable exception) {
            return 0;
        }
    }

}
