/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.ws.util;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.IOUtils;
import org.glassfish.jersey.message.MessageUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.http.HttpServletResponse;
import javax.ws.rs.BadRequestException;
import javax.ws.rs.ClientErrorException;
import javax.ws.rs.ForbiddenException;
import javax.ws.rs.InternalServerErrorException;
import javax.ws.rs.NotAcceptableException;
import javax.ws.rs.NotAllowedException;
import javax.ws.rs.NotAuthorizedException;
import javax.ws.rs.NotFoundException;
import javax.ws.rs.NotSupportedException;
import javax.ws.rs.ServiceUnavailableException;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.client.ClientRequestContext;
import javax.ws.rs.client.ClientRequestFilter;
import javax.ws.rs.client.ClientResponseContext;
import javax.ws.rs.client.ClientResponseFilter;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.InputStreamReader;
import java.nio.charset.Charset;

import static javax.servlet.http.HttpServletResponse.SC_METHOD_NOT_ALLOWED;

/**
 * A filter used to handle 40x and 50x series errors. These extract any error messages returned in the body of
 * responses, and includes them in the thrown exception.
 * <p/>
 * This supports both JSON and plain text responses. <p/>
 * For JSON responses, a constructor is provided that accepts a class which is used to deserialise the error
 * message. This must provide a {@code toString()} implementation that returns the message.
 *
 * @author Tim Anderson
 */
public class ErrorResponseFilter implements ClientResponseFilter {

    /**
     * The object mapper.
     */
    private final ObjectMapper mapper;

    /**
     * The class responsible for deserialising errors, if the response is JSON.
     */
    private final Class<?> errorType;

    /**
     * The logger.
     */
    private static final Logger log = LoggerFactory.getLogger(ErrorResponseFilter.class);

    /**
     * Constructs an {@link ErrorResponseFilter}.
     *
     * @param mapper the mapper
     */
    public ErrorResponseFilter(ObjectMapper mapper) {
        this(mapper, null);
    }

    /**
     * Constructs an {@link ErrorResponseFilter}.
     *
     * @param mapper    the mapper
     * @param errorType the class responsible for deserialising errors, for JSON responses
     */
    public ErrorResponseFilter(ObjectMapper mapper, Class<?> errorType) {
        this.mapper = mapper;
        this.errorType = errorType;
    }

    /**
     * Filter method called after a response has been provided for a request (either by a
     * {@link ClientRequestFilter request filter} or when the HTTP invocation returns.
     * <p>
     * Filters in the filter chain are ordered according to their {@code javax.annotation.Priority}
     * class-level annotation value.
     *
     * @param requestContext  request context.
     * @param responseContext response context.
     */
    @Override
    public void filter(ClientRequestContext requestContext, ClientResponseContext responseContext) {
        Response.StatusType statusInfo = responseContext.getStatusInfo();
        if (statusInfo.getFamily() == Response.Status.Family.CLIENT_ERROR
            || statusInfo.getFamily() == Response.Status.Family.SERVER_ERROR) {
            MediaType mediaType = responseContext.getMediaType();
            if (responseContext.hasEntity() && MediaTypes.isA(mediaType, MediaType.APPLICATION_JSON_TYPE,
                                                              MediaType.TEXT_PLAIN_TYPE)) {
                String message = getErrorMessage(responseContext, mediaType);

                WebApplicationException exception;
                int status = responseContext.getStatus();
                switch (status) {
                    case HttpServletResponse.SC_BAD_REQUEST:
                        exception = new BadRequestException(message);
                        break;
                    case HttpServletResponse.SC_UNAUTHORIZED:
                        exception = new NotAuthorizedException(message, Response.status(status).build());
                        break;
                    case HttpServletResponse.SC_FORBIDDEN:
                        exception = new ForbiddenException(message);
                        break;
                    case HttpServletResponse.SC_NOT_FOUND:
                        exception = new NotFoundException(message);
                        break;
                    case SC_METHOD_NOT_ALLOWED:
                        exception = createNotAllowedException(message, responseContext);
                        break;
                    case HttpServletResponse.SC_NOT_ACCEPTABLE:
                        exception = new NotAcceptableException(message);
                        break;
                    case HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE:
                        exception = new NotSupportedException(message);
                        break;
                    case HttpServletResponse.SC_INTERNAL_SERVER_ERROR:
                        exception = new InternalServerErrorException(message);
                        break;
                    case HttpServletResponse.SC_SERVICE_UNAVAILABLE:
                        exception = new ServiceUnavailableException(message);
                        break;
                    default:
                        // For servers that uses custom status. E.g. SFS uses 465 Access to the Document Denied
                        exception = new WebApplicationException(message, status);
                }

                throw exception;
            }
        }
    }

    /**
     * Returns an exception for {@code SC_METHOD_NOT_ALLOWED}.
     * <p/>
     *
     * @param message         the error message
     * @param responseContext the response context
     * @return the exception
     */
    private WebApplicationException createNotAllowedException(String message, ClientResponseContext responseContext) {
        WebApplicationException result = null;
        try {
            String allowed = responseContext.getHeaderString(HttpHeaders.ALLOW);
            Response response = Response.status(SC_METHOD_NOT_ALLOWED).header(HttpHeaders.ALLOW, allowed).build();
            result = new NotAllowedException(message, response);
        } catch (Throwable exception) {
            log.error("Failed to create NotAllowedException: " + exception.getMessage(), exception);
        }
        if (result == null) {
            // fall back to ClientErrorException
            return new ClientErrorException(message, SC_METHOD_NOT_ALLOWED);
        }
        return result;
    }

    /**
     * Returns the error message from the message body.
     *
     * @param responseContext the response context
     * @param mediaType       the media type
     * @return the error message. May be {@code null}
     */
    private String getErrorMessage(ClientResponseContext responseContext, MediaType mediaType) {
        String result = null;
        try {
            if (errorType != null && MediaTypes.isA(mediaType, MediaType.APPLICATION_JSON_TYPE)) {
                Object error = mapper.readValue(responseContext.getEntityStream(), errorType);
                result = (error != null) ? error.toString() : null;
            } else {
                Charset charset = MessageUtils.getCharset(mediaType);
                InputStreamReader reader = new InputStreamReader(responseContext.getEntityStream(), charset);
                result = IOUtils.toString(reader);
            }
        } catch (Throwable exception) {
            log.error("Failed to deserialise error from response: " + exception.getMessage(), exception);
        }
        return result;
    }

}